# -*- coding: utf-8 -*-
import unittest

from nidaba import nidaba


class BatchTest(unittest.TestCase):

    """Tests the functionality of the Batch class."""

    def test_no_tick(self):
        pass

    def test_no_step(self):
        pass

    def test_no_task(self):
        pass

    def test_no_doc(self):
        pass

    def test_multiple_docs(self):
        pass

    def test_multiple_tasks(self):
        pass

    def test_multiple_ticks(self):
        pass

    def test_multiple_steps(self):
        pass

if __name__ == '__main__':
    unittest.main()
